""" CLI for manipulating symbol maps """
import click
from .config import Config, ConfigKeys
from .dal import get_session, SymbolMap

@click.group("map")
def symbol_map():
    """ Operations with symbol maps """
    pass

@click.command("add")
@click.argument("incoming") # , "in"
@click.argument("outgoing") # , "out"
def add_map(incoming, outgoing):
    """ Creates a symbol mapping """
    db_path = Config().get(ConfigKeys.pricedb_path)
    session = get_session(db_path)

    new_map = SymbolMap()
    new_map.in_symbol = incoming
    new_map.out_symbol = outgoing

    session.add(new_map)
    session.commit()
    click.echo("Record saved.")

@click.command("list")
def list_maps():
    """ Displays all symbol maps """
    db_path = Config().get(ConfigKeys.price_database)
    session = get_session(db_path)

    maps = session.query(SymbolMap).all()
    for item in maps:
        click.echo(item)

###
symbol_map.add_command(add_map)
symbol_map.add_command(list_maps)



"""
Alphavantage downloader
"""
import logging
from datetime import datetime
from decimal import Decimal

from pricedb.config import Config, ConfigKeys
from pricedb.model import PriceModel, SecuritySymbol


class AlphaVantageDownloader:
    """ Uses AlphaVantage to get prices """
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        # Provide the api key value.
        self.api_key = None

    def download(self, symbol: SecuritySymbol, currency: str):
        """ Download price """
        from alpha_vantage.timeseries import TimeSeries

        cfg = Config()
        self.api_key = cfg.get(ConfigKeys.alphavantage_api_key)
        ts = TimeSeries(key=self.api_key)

        # pylint: disable=E0632
        # data, meta_data = ts.get_daily(symbol)
        data, meta_data = ts.get_intraday(symbol.mnemonic)
        # pylint: enable=E0632

        # pylint: disable=no-member
        keys = list(data.keys())
        # pylint: enable=no-member
        latest_key = keys[0]
        latest_price = data[latest_key]
        value = latest_price["4. close"].strip("0")

        # Parse
        result = PriceModel()
        result.symbol = symbol
        result.datetime = datetime.strptime(latest_key, "%Y-%m-%d %H:%M:%S")
        result.value = Decimal(value)
        result.currency = currency

        self.logger.debug(f"{latest_key}, {result}")
        return result



""" Various utilities """
from typing import List


def read_lines_from_file(file_path: str) -> List[str]:
    """ Read text lines from a file """
    # check if the file exists?
    with open(file_path) as csv_file:
        content = csv_file.readlines()
    return content



"""
Configuration handling
The config file is stored in data directory and is located by using Resources.
Ref: http://peak.telecommunity.com/DevCenter/setuptools#non-package-data-files
"""
import io
import os.path
import shutil
from configparser import ConfigParser
from enum import Enum, auto
import logging

from pkg_resources import Requirement, resource_filename

package_name = "PriceDb"
template_path = "pricedb/templates/"
config_filename = "pricedb.ini"
SECTION = "Default"


class ConfigKeys(Enum):
    price_database = auto(),
    alphavantage_api_key = auto()
    fixerio_api_key = auto()


class Config:
    """ Reads and writes Asset Allocation config file """

    def __init__(self, ini_path: str = None):
        self.logger = logging.getLogger(__name__)

        # Read the config file on creation of the object.
        self.config = ConfigParser()

        if not ini_path:
            # use the default path.
            file_path = os.path.abspath(self.get_config_path())
            # copy the template if file does not yet exist
            if not os.path.exists(file_path):
                self.__create_user_config()
        else:
            file_path = os.path.abspath(ini_path)

        self.__read_config(file_path)

    def delete_user_config(self):
        """ Delete current user's config file """
        file = self.get_config_path()
        os.remove(file)

    def __read_config(self, file_path: str):
        """ Read the config file """
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File path not found: {file_path}")
        # check if file exists
        if not os.path.isfile(file_path):
            self.logger.error(f"file not found: {file_path}")
            raise FileNotFoundError(f"configuration file not found {file_path}")

        self.config.read(file_path)

    def __get_config_template_path(self) -> str:
        """ gets the default config path from resources """
        filename = resource_filename(
            Requirement.parse(package_name),
            template_path + config_filename)
        return filename

    def __get_user_path(self) -> str:
        """ Returns the current user's home directory """
        return os.path.expanduser("~")

    def __create_user_config(self):
        """ Copy the config template into user's directory """
        src_path = self.__get_config_template_path()
        src = os.path.abspath(src_path)

        if not os.path.exists(src):
            message = f"Config template not found {src}"
            self.logger.error(message)
            raise FileNotFoundError(message)

        dst = os.path.abspath(self.get_config_path())

        shutil.copyfile(src, dst)

        if not os.path.exists(dst):
            raise FileNotFoundError("Config file could not be copied to user dir!")

    def get_config_path(self) -> str:
        """
        Returns the path where the active config file is expected.
        This is the user's profile folder.
        """
        dst_dir = self.__get_user_path()
        dst = dst_dir + "/" + config_filename
        return dst

    def get_contents(self) -> str:
        """ Reads the contents of the config file """
        content = None
        # with open(file_path) as cfg_file:
        #     contents = cfg_file.read()

        # Dump the current contents into an in-memory file.
        in_memory = io.StringIO("")
        self.config.write(in_memory)
        in_memory.seek(0)
        content = in_memory.read()
        #     log(DEBUG, "config content: %s", content)
        in_memory.close()
        return content

    def set(self, option: ConfigKeys, value):
        """ Sets a value in config """
        assert isinstance(option, ConfigKeys)

        # As currently we only have 1 section.
        section = SECTION
        self.config.set(section, option.name, value)
        self.save()

    def get(self, option: ConfigKeys):
        """ Retrieves a config value """
        assert isinstance(option, ConfigKeys)

        # Currently only one section is used
        section = SECTION
        return self.config.get(section, option.name)

    def save(self):
        """ Save the config file """
        file_path = self.get_config_path()
        contents = self.get_contents()
        with open(file_path, mode='w') as cfg_file:
            cfg_file.write(contents)



"""
Fixer.io currency rates
Currently only implements the free public API access with yesterday's rates.
"""
import os
import tempfile
import logging
from decimal import Decimal
from pydatum import Datum

try:
    import simplejson as json
except ImportError:
    import json
from pricedb import SecuritySymbol
from pricedb.config import Config, ConfigKeys


class Quote:
    """Class to represent a quote (price, ...)"""

    def __init__(self):
        self.datum: Datum = Datum()
        #self.namespace: str = None
        self.symbol: SecuritySymbol = None
        self.value: Decimal = Decimal(0)
        self.currency: str = None

    def __repr__(self):
        # symbol = ("{namespace}:{symbol}".format(namespace=self.namespace, symbol=self.symbol)
        #           if self.namespace else self.symbol)
        symbol = repr(self.symbol)
        symbol = "{symbol:<13}".format(symbol=symbol)

        value = "{value:>6}".format(value=self.value)
        return f"<Quote ('{symbol}',date:{self.datum},value:{value},currency:{self.currency})>"


class FixerioQuote(Quote):
    """ Fixer.io-specific quote """
    pass


class Fixerio:
    """Retrieves prices from data files or online provider(s)"""

    def __init__(self):
        self.cache_path = tempfile.gettempdir()
        self.logger = logging.getLogger(__name__)

        # read fixer.io api key
        cfg = Config()
        self.api_key = cfg.get(ConfigKeys.fixerio_api_key)

    def download(self, symbol: SecuritySymbol, currency: str) -> FixerioQuote:
        """ Download latest rates. Caches into temp directory. """
        if symbol.namespace:
            symbol.namespace = symbol.namespace.upper()
            # assert namespace == "CURRENCY"
        # namespace is ignored, anyways.
        currency = currency.upper()
        symbol.mnemonic = symbol.mnemonic.upper()
        # make sure the symbol does not contain namespace
        if ":" in symbol.mnemonic:
            raise ValueError("Currency symbol should not contain namespace.")

        rates_dict = None
        if self.latest_rates_exist():
            self.logger.debug("Cached rates found.")
            rates_dict = self.__read_rates_from_file()
        else:
            rates_dict = self.__download_rates(currency)

        mapper = FixerioModelMapper(rates_dict)
        model = mapper.get_model_for_symbol(symbol.mnemonic)
        self.logger.debug(f"current model: {model}")
        return model

    def __download_rates(self, base_currency: str):
        """
        Downloads the latest rates. Requires base currency and a list of currencies to
        retrieve.
        Returns json response object from Fixer.io.
        """
        import requests

        assert isinstance(base_currency, str)

        # Downloads the latest rates using Fixerio. Returns dict.
        # https://pypi.python.org/pypi/fixerio
        # from fixerio import Fixerio
        # fxrio = Fixerio(base=base_currency) #, symbols=symbols)
        # latest_rates = fxrio.latest()

        result = None
        base_url = 'http://data.fixer.io/api/latest'
        query = f"{base_url}?base={base_currency}&access_key={self.api_key}"

        # if symbols:
        #     symbols_csv = ",".join(symbols)
        # if symbols_csv:
        #     query += f"&symbols={symbols_csv}"

        try:
            self.logger.debug(f"retrieving rates from {query}")
            response = requests.get(query)
            # print("[%s] %s" % (response.status_code, response.url))
            if response.status_code != 200:
                result = 'N/A'
            else:
                result = response.json()
                # rate_in_currency = rates["rates"][rate_in]
        except requests.ConnectionError as error:
            self.logger.error(error)

        self.logger.debug("Latest prices downloaded.")

        # Since these are daily rates, cache them into a file.
        # Ignored for now since we are downloading individual currencies, not all together.
        self.__save_rates(result)

        return result

    def get_yesterdays_file_path(self):
        """ Full path to the today's rates file. """
        datum = Datum()
        datum.yesterday()
        yesterday = datum.get_iso_date_string()

        return self.__get_rate_file_path(yesterday)

    def get_todays_file_path(self):
        """ path to today's cached file """
        datum = Datum()
        datum.today()
        today = datum.get_iso_date_string()

        return self.__get_rate_file_path(today)

    def latest_rates_exist(self):
        """ Check if latest rates cached file exists. """
        file_path = self.get_todays_file_path()
        exists = os.path.isfile(file_path)

        if exists:
            self.logger.debug(f"Cached file found at {file_path}")

        return exists

    def __get_rate_file_path(self, filename):
        """
        Assemble full file path for the given name (date).
        """
        return os.path.abspath("{cache_path}/fixerio_{filename}.json".format(
            cache_path=self.cache_path, filename=filename))

    def __read_rates_from_file(self):
        file_path = self.get_todays_file_path()

        with open(file_path, 'r') as file:
            content = file.read()

        # self.logger.debug(f"cached rates: {content}")
        return json.loads(content)

    def __save_rates(self, rates):
        """
        Saves the retrieved rates into a cache file
        """
        file_date = rates["date"]
        filename = self.__get_rate_file_path(file_date)

        content = json.dumps(rates)

        with open(filename, 'w') as file:
            file.write(content)
            self.logger.debug(f"rates saved in {filename}.")


class FixerioModelMapper:
    """ Maps the result from Fixer.io into an array of FixerioQuote """

    def __init__(self, json_response):
        self.__data = json_response

    def get_model_for_symbol(self, symbol: str) -> FixerioQuote:
        """ Read and map a single currency rate """
        date_str = self.__data["date"]
        # rate_date = datetime.strptime(date_str, "%Y-%m-%d")
        # assert isinstance(rate_date, datetime)

        base = self.__data["base"]
        rates_dict = self.__data["rates"]
        value_str = rates_dict[symbol]
        value = Decimal(value_str)
        # The rate is inverse value.
        rate = Decimal(1) / value
        # Round to 6 decimals max.
        rounded_str = "{0:.6f}".format(rate)
        rounded = Decimal(rounded_str)

        model = Quote()
        model.symbol = SecuritySymbol("CURRENCY", symbol)
        model.value = rounded
        model.datum.from_iso_date_string(date_str)
        model.currency = base

        return model



""" Domain model. Public """
from decimal import Decimal
from datetime import datetime
from pydatum import Datum


class SecuritySymbol:
    """ Symbol security, using namespace and the mnemonic """
    def __init__(self, namespace: str, mnemonic: str):
        self.namespace = namespace
        self.mnemonic = mnemonic

    def __repr__(self):
        if self.namespace:
            return f"{self.namespace}:{self.mnemonic}"
        else:
            return self.mnemonic

    def parse(self, symbol: str) -> (str, str):
        """ Splits the symbol into namespace, symbol tuple """
        symbol_parts = symbol.split(":")
        namespace = None
        mnemonic = symbol

        if len(symbol_parts) > 1:
            namespace = symbol_parts[0]
            mnemonic = symbol_parts[1]

        self.namespace = namespace
        self.mnemonic = mnemonic

        return namespace, mnemonic


class PriceModel:
    """ The price element """
    def __init__(self):
        # self.datetime: datetime = None
        self.datum: Datum = Datum()
        self.symbol: SecuritySymbol = SecuritySymbol("", "")
        self.value: Decimal = Decimal(0)
        self.currency: str = None

    def __repr__(self):
        symbol = f"{str(self.symbol):<13}"

        value = f"{self.value:>6}"
        return f"<Price ('{symbol}', date:{self.datum}, value:{value}, currency:{self.currency})>"



""" Repositories - operations on multiple entities/aggregates """
from typing import List
from .dal import SymbolMap, Price, get_default_session, Security


class SymbolMapRepository:
    """ Operations on SymbolMap collections """
    def __init__(self, session):
        self.session = session

    def get_by_id(self, symbol: str) -> SymbolMap:
        """ Finds the map by in-symbol """
        return self.query.filter(SymbolMap.in_symbol == symbol).first()

    def get_all(self) -> List[SymbolMap]:
        """ Returns all maps """
        return self.query.all()

    @property
    def query(self):
        """ Basic query """
        return self.session.query(SymbolMap)


class PriceRepository:
    """ Operations on prices """
    def __init__(self, session=None):
        self.__session = session

    def add(self, entity: Price):
        """ Insert price """
        self.session.add(entity)

    def save(self):
        """ Save all changes in the session """
        self.session.commit()

    @property
    def session(self):
        """ db session """
        if not self.__session:
            self.__session = get_default_session()
        return self.__session

    @property
    def query(self):
        """ Base query """
        return self.session.query(Price)


class SecurityRepository:
    """ Security table """
    def __init__(self, session=None):
        self.__session = session

    @property
    def session(self):
        """ db session """
        if not self.__session:
            self.__session = get_default_session()
        return self.__session

    @property
    def query(self):
        """ Base query """
        return self.session.query(Security)



""" Morningstar price downloader """
import logging
import urllib.parse
import urllib.request
from datetime import datetime
from decimal import Decimal, InvalidOperation
from pydatum import Datum

from bs4 import BeautifulSoup

from pricedb.model import PriceModel, SecuritySymbol


class MorningstarDownloader:
    """ Fetches prices from Morningstar site """
    def __init__(self):
        self.url = "http://quotes.morningstar.com/stockq/c-header"
        self.params = {'t': 'symbol'}
        self.namespaces = {
            "AMS": "XAMS",
            "ASX": "XASX",
            "XETRA": "XETR",
            "LSE": "XLON",
            "NASDAQ": "XNAS",
            "NYSE": "XNYS",
            "NYSEARCA": "ARCX"
        }
        self.logger = logging.getLogger(__name__)

    def download(self, symbol: SecuritySymbol, currency: str) -> PriceModel:
        """ Download the given symbol """
        if not symbol.namespace:
            raise ValueError(f"Namespace not sent for {symbol}")

        # get the translated namespace
        local_namespace = self.namespaces[symbol.namespace]
        self.params["t"] = f"{local_namespace}:{symbol.mnemonic}"
        # assemble url
        params = urllib.parse.urlencode(self.params)
        url = self.url + '?' + params
        self.logger.debug(f"fetching price from {url}")

        with urllib.request.urlopen(url) as response:
            html = response.read()

        if not html:
            return None

        # parse HTML
        price = self.parse_price(html)
        if price:
            price.symbol = symbol
        # compare currency
        if price.currency != currency:
            raise ValueError(f"Requested currency ({currency}) does not match the {symbol} -> {currency}.")

        return price

    def parse_price(self, page: str) -> PriceModel:
        """ parse html to get the price """
        result = PriceModel()
        soup = BeautifulSoup(page, 'html.parser')

        # Price value
        price_el = soup.find(id='last-price-value')
        if not price_el:
            logging.debug(f"Received from mstar: {page}")
            raise ValueError("No price info found in returned HTML.")

        value = price_el.get_text().strip()
        try:
            result.value = Decimal(value)
        except InvalidOperation:
            message = f"Could not parse price value {value}"
            print(message)
            self.logger.error(message)
            return None

        # The rest
        date_str = soup.find(id="asOfDate").get_text().strip()
        date_val = datetime.strptime(date_str, "%m/%d/%Y %H:%M:%S")
        result.datum = Datum()
        result.datum.from_datetime(date_val)

        # tz_str = soup.find(id="timezone").get_text().strip()

        currency = soup.find(id="curency").get_text().strip()
        result.currency = currency

        return result
